# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

from typing import TypedDict


class ProviderConfig(TypedDict):
    temperature: float


class Provider(TypedDict):
    id: str
    config: ProviderConfig



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Callable
from typing import Any, Literal, Optional, TypedDict

from langchain.prompts.prompt import PromptTemplate
from langsmith.evaluation import (
    EvaluationResult,
    EvaluationResults,
    LangChainStringEvaluator,
)
from langsmith.schemas import Example, Run

from langsmith_evaluation_helper.llm.model import ChatModel, ChatModelName


class BuiltinEvaluatorConfig(TypedDict):
    type: Literal["length", "llm-judge", "similar"]
    value: str
    label: Optional[str]  # noqa: UP007
    judge_provider: dict[Any, Any] | None


class EvalResult(TypedDict):
    key: str | None
    score: float


Evaluator = Callable[[Run, Example], EvalResult | EvaluationResult | EvaluationResults]


def create_length_evaluator(evaluator_config: BuiltinEvaluatorConfig) -> Evaluator:
    def length_evaluator(run: Run, example: Example) -> EvalResult:
        output = run.outputs.get("output") if run.outputs is not None else None
        if output is None:
            return {"key": "length", "score": False}

        value = evaluator_config["value"]
        score = False

        try:
            if "<=" in value:
                score = len(output) <= int(value.replace("<=", "").strip())
            elif "<" in value:
                score = len(output) < int(value.replace("<", "").strip())
            elif ">=" in value:
                score = len(output) >= int(value.replace(">=", "").strip())
            elif ">" in value:
                score = len(output) > int(value.replace(">", "").strip())
            else:
                raise ValueError
        except Exception:
            print(f"[Warning] Invalid integer value in evaluator_config['value']: {value}")
            return {"key": "length", "score": False}

        return {"key": "length", "score": score}

    return length_evaluator


def create_llm_judge_evaluator(evaluator_config: BuiltinEvaluatorConfig) -> Evaluator:
    def llm_judge_evaluator(run: Run, example: Example) -> EvalResult:
        judge_provider = evaluator_config.get("judge_provider")
        if judge_provider is None or ("id" not in judge_provider) or ("config" not in judge_provider):
            raise ValueError("Please provide llm-judge judge_provider id and config")
        judge_model_id = judge_provider["id"]
        judge_model = getattr(ChatModelName, judge_model_id, None)

        temperature = judge_provider["config"]["temperature"]
        if judge_model is not None:
            model = ChatModel(default_model_name=judge_model, temperature=temperature, verbose=True)
        else:
            raise ValueError(f"Invalid judge model_id: {judge_model_id}")
        evaluate_prompt = """Evaluate and give a score between 0 to 1 the following text with the evaluation perspective specified in the prompt.

          evaluation perspective: {evaluation_perspective}
          text: {text}

          only output the score
          """

        prompt = PromptTemplate.from_template(evaluate_prompt)
        inputs = {
            "evaluation_perspective": evaluator_config["value"],
            "text": run.outputs.get("output") if run.outputs is not None else None,
        }

        key = evaluator_config.get("label", "llm-judge")
        score = float(model.invoke(prompt=prompt, **inputs))  # type: ignore[arg-type]
        return {"key": key, "score": score}

    return llm_judge_evaluator


def create_similar_evaluator() -> Evaluator:
    def similar_evaluator(run: Run, example: Example) -> EvaluationResult | EvaluationResults:
        evaluator = LangChainStringEvaluator("embedding_distance")
        run_evaluator = evaluator.as_run_evaluator()
        evaluation_result = run_evaluator.evaluate_run(run, example)
        return evaluation_result

    return similar_evaluator


def generate_builtin_evaluator_functions(
    evaluator_configs: list[BuiltinEvaluatorConfig],
) -> list[Evaluator]:
    evaluators = []

    for evaluator_config in evaluator_configs:
        if evaluator_config["type"] == "length":
            evaluators.append(create_length_evaluator(evaluator_config))
        elif evaluator_config["type"] == "llm-judge":
            evaluators.append(create_llm_judge_evaluator(evaluator_config))
        elif evaluator_config["type"] == "similar":
            evaluators.append(create_similar_evaluator())
        else:
            print(f"[Warning] Unknown evaluator type: {evaluator_config['type']}")

    return evaluators



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

import os
from enum import Enum
from typing import Any

from langchain.chat_models.base import BaseChatModel
from langchain.schema.output_parser import StrOutputParser
from langchain_anthropic import ChatAnthropic
from langchain_core.prompts import PromptTemplate
from langchain_core.pydantic_v1 import SecretStr
from langchain_core.runnables import RunnableConfig
from langchain_google_vertexai import ChatVertexAI
from langchain_openai import AzureChatOpenAI, ChatOpenAI


class ChatModelName(Enum):
    TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4-0613"
    GPT4_32K = "gpt-4-32k-0613"
    GPT4O = "gpt-4o"
    AZURE_GPT35_16K_TURBO = "gpt-35-turbo"
    AZURE_GPT4_32K = "gpt-4-32k"
    GEMINI_PRO = "gemini-pro"
    GEMINI_FLASH = "gemini-1.5-flash-001"
    CLAUDE3_SONNET = "claude3-sonnet"
    CLAUDE3_OPUS = "claude-3-opus-20240229"
    CLAUDE3_HAIKU = "claude-3-haiku-20240307"
    CLAUDE3_5_SONNET = "claude-3-5-sonnet-20240620"


AZURE_OPENAI_API_KEY = os.getenv("AZURE_OPENAI_API_KEY", "")
AZURE_OPENAI_API_BASE = os.getenv("AZURE_OPENAI_API_BASE", "")


def get_chat_model(name: ChatModelName, **kwargs: Any) -> BaseChatModel:
    azure_deployment = kwargs.pop("azure_deployment") if "azure_deployment" in kwargs else None
    api_version = kwargs.pop("api_version") if "api_version" in kwargs else None
    if (
        name == ChatModelName.TURBO
        or name == ChatModelName.GPT4
        or name == ChatModelName.GPT4_32K
        or name == ChatModelName.GPT4O
    ):
        return ChatOpenAI(model=name.value, **kwargs)
    elif (
        name == ChatModelName.AZURE_GPT35_16K_TURBO
        and ((azure_deployment is not None) and (api_version is not None))
        or name == ChatModelName.AZURE_GPT4_32K
        and ((azure_deployment is not None) and (api_version is not None))
    ):
        return AzureChatOpenAI(
            api_key=SecretStr(AZURE_OPENAI_API_KEY),
            azure_endpoint=AZURE_OPENAI_API_BASE,
            api_version=api_version,
            azure_deployment=azure_deployment,
            **kwargs,
        )
    elif name == ChatModelName.GEMINI_PRO or name == ChatModelName.GEMINI_FLASH:
        return ChatVertexAI(
            model_name=name.value,
            **kwargs,
        )
    elif (
        name == ChatModelName.CLAUDE3_SONNET
        or name == ChatModelName.CLAUDE3_OPUS
        or name == ChatModelName.CLAUDE3_HAIKU
        or name == ChatModelName.CLAUDE3_5_SONNET
    ):
        return ChatAnthropic(model_name="claude-3-sonnet-20240229", **kwargs)
    else:
        raise ValueError(f"Invalid model name. {name}")


class ChatModel:
    """Facade wrapper class to handle lifecycle of Langchain's basechatmodel."""

    default_model: BaseChatModel
    default_model_name: ChatModelName
    verbose: bool = True
    kwargs: Any

    def __init__(
        self,
        default_model_name: ChatModelName = ChatModelName.CLAUDE3_SONNET,
        **kwargs: Any,
    ) -> None:
        self.default_model = get_chat_model(default_model_name, **kwargs)
        self.default_model_name = default_model_name
        self.kwargs = kwargs

    def get_model(self, model_name: ChatModelName | None = None) -> BaseChatModel:
        model = self.default_model
        if model_name:
            model = get_chat_model(model_name, **self.kwargs)

        return model

    def invoke(
        self,
        prompt: PromptTemplate,
        tags: list[str] | None = None,
        model_name: ChatModelName | None = None,
        metadata: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> str:
        chain = (prompt | self.get_model(model_name)) | StrOutputParser()
        config = RunnableConfig(
            tags=tags or [],
            metadata=metadata or {},
        )

        return chain.invoke(input=kwargs, config=config)

    async def async_invoke(
        self,
        prompt: PromptTemplate,
        tags: list[str] | None = None,
        model_name: ChatModelName | None = None,
        metadata: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> str:
        chain = prompt | self.get_model(model_name) | StrOutputParser()
        config = RunnableConfig(
            tags=tags or [],
            metadata=metadata or {},
        )

        return await chain.ainvoke(input=kwargs, config=config)



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

import inspect
import os
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any

from langchain.prompts import PromptTemplate
from langsmith import traceable

from langsmith_evaluation_helper.llm.model import ChatModel, ChatModelName
from langsmith_evaluation_helper.llm.prompt_template_wrapper import (
    InputTypedPromptTemplate,
)
from langsmith_evaluation_helper.utils import is_async_function, load_function


def load_prompt(config_path: str, config: dict[Any, Any]) -> Any:
    """
    Load the config file and execute the specified entry function.
    """
    prompt_info = config["prompt"]
    script_path = os.path.join(os.path.dirname(config_path), prompt_info["name"])
    function_name = prompt_info["entry_function"]

    func = load_function(script_path, function_name)

    return func


def has_inputs_argument(func: Callable[..., Any]) -> bool:
    signature = inspect.signature(func)
    parameters = signature.parameters
    return "inputs" in parameters


def get_prompt_template_and_kwargs_from_input_typed_prompt_template(
    prompt: InputTypedPromptTemplate,
) -> tuple[PromptTemplate, dict[str, Any]]:
    return PromptTemplate.from_template(prompt.template), prompt.input.model_dump()


def get_prompt_template_and_kwargs_from_inputs(
    prompt: str, inputs: dict[Any, Any]
) -> tuple[PromptTemplate, dict[str, Any]]:
    kwargs: dict[str, Any] = {}
    for key, _ in inputs.items():
        kwargs[key] = inputs.get(key, "Unknown")

    return PromptTemplate.from_template(prompt), kwargs


def has_input_typed_prompt_template_properties(obj: Any) -> bool:
    return hasattr(obj, "template") and hasattr(obj, "input")


@traceable
def execute_prompt(
    inputs: dict[Any, Any],
    prompt: str | InputTypedPromptTemplate,
    provider: dict[Any, Any],
) -> str:
    if isinstance(prompt, str):
        _prompt_template, kwargs = get_prompt_template_and_kwargs_from_inputs(prompt, inputs)
    elif has_input_typed_prompt_template_properties(prompt):
        _prompt_template, kwargs = get_prompt_template_and_kwargs_from_input_typed_prompt_template(prompt)
    else:
        raise ValueError(f"Invalid prompt type: {type(prompt)}")

    model_id = provider["id"]
    model = getattr(ChatModelName, model_id, None)
    provider_config = provider.get("config", {})
    temperature = provider_config.get("temperature", 0)
    azure_deployment = provider_config.get("azure_deployment", None)
    azure_api_version = provider_config.get("azure_api_version", None)
    if ("AZURE" in model_id) and (azure_deployment is None or azure_api_version is None):
        raise ValueError("Add azure_deployment and azure_api_version to config for Azure GPT models")
    if model is not None:
        llm = ChatModel(
            default_model_name=model,
            temperature=temperature,
            azure_deployment=azure_deployment,
            api_version=azure_api_version,
            verbose=True,
        )
    else:
        raise ValueError(f"Invalid model_id: {model_id}")

    result = llm.invoke(_prompt_template, **kwargs)
    return result


def load_prompt_template(
    config_path: str, prompt_config: dict[Any, Any], provider: dict[Any, Any]
) -> Callable[[dict[str, Any]], str] | Callable[[dict[str, Any]], Awaitable[str]]:
    prompt_func = load_prompt(config_path, prompt_config)
    is_async = is_async_function(prompt_func)
    has_inputs = has_inputs_argument(prompt_func)

    async def run_async(inputs: dict[str, Any]) -> str:
        prompt = await prompt_func(inputs) if has_inputs else await prompt_func()
        return execute_prompt(inputs, prompt, provider)

    def run_sync(inputs: dict[str, Any]) -> str:
        prompt = prompt_func(inputs) if has_inputs else prompt_func()
        return execute_prompt(inputs, prompt, provider)

    return run_async if is_async else run_sync


def load_prompt_function(
    config_path: str,
    prompt_file_name: str,
    entry_function_name: str,
    provider: dict[Any, Any],
) -> Callable[[dict[str, Any]], Any] | Callable[[dict[str, Any]], Coroutine[Any, Any, Any]]:
    script_path = os.path.join(os.path.dirname(config_path), prompt_file_name)
    function_name = entry_function_name
    prompt_func = load_function(script_path, function_name)
    is_async = is_async_function(prompt_func)
    params = inspect.signature(prompt_func).parameters

    def create_kwargs(inputs: dict[Any, Any]) -> dict[str, Any]:
        kwargs = {}
        if "inputs" in params:
            kwargs["inputs"] = inputs
        if "provider" in params:
            kwargs["provider"] = provider
        return kwargs

    if is_async:

        async def arun(inputs: dict[Any, Any]) -> Any:
            return await prompt_func(**create_kwargs(inputs))

        return arun

    def run(inputs: dict[Any, Any]) -> Any:
        return prompt_func(**create_kwargs(inputs))

    return run


def load_run_function(
    config_path: str, config: dict[Any, Any], provider: dict[Any, Any]
) -> Callable[[dict[Any, Any]], Any]:
    prompt_config = config.get("prompt")
    custom_run_config = config.get("custom_run")

    if prompt_config is not None and custom_run_config is not None:
        raise ValueError("prompt and custom_run cannot be specified simultaneously")

    if custom_run_config:
        return load_prompt_function(
            config_path,
            custom_run_config["file_path"],
            custom_run_config["entry_function"],
            provider,
        )
    else:
        return load_prompt_template(config_path, config, provider)



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
import sys

from dotenv import load_dotenv

load_dotenv()


def evaluate(config_path: str) -> None:
    # Determine the absolute path of loader.py within the package
    package_dir = os.path.dirname(os.path.realpath(__file__))
    loader_path = os.path.join(package_dir, "loader.py")

    if not os.path.isfile(loader_path):
        print(f"Error: loader.py not found at {loader_path}")
        sys.exit(1)

    subprocess.run([sys.executable, loader_path, config_path])


def main() -> None:
    if len(sys.argv) < 2:
        print("Usage: langsmith-evaluation-helper evaluate /path/to/config")
        sys.exit(1)

    command = sys.argv[1]

    if command == "evaluate":
        if len(sys.argv) != 3:
            print("Usage: langsmith-evaluation-helper evaluate /path/to/config")
            sys.exit(1)
        config_path = sys.argv[2]
        evaluate(config_path)
    else:
        print(f"Unknown command: {command}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

from typing import Generic, TypeVar

from pydantic import BaseModel

PromptInput = TypeVar("PromptInput", bound=BaseModel)


class InputTypedPromptTemplate(BaseModel, Generic[PromptInput]):
    template: str
    input: PromptInput



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import os
import sys
from typing import Any

import yaml
from langsmith import Client, aevaluate, evaluate

from langsmith_evaluation_helper.builtin_evaluators import (
    generate_builtin_evaluator_functions,
)
from langsmith_evaluation_helper.load_run_function import load_run_function
from langsmith_evaluation_helper.utils import is_async_function, load_function

LANGCHAIN_TENANT_ID = os.getenv("LANGCHAIN_TENANT_ID", None)
MAX_EXAMPLES_COUNT = 1000


def load_config(config_path: Any) -> dict[str, Any]:
    """
    Loads config file as YML dictionary.
    """
    with open(config_path) as file:
        config = yaml.safe_load(file)
    return config


def load_dataset(config: dict[Any, Any]) -> tuple[Any, Any, Any, list[str]]:
    test_info = config["tests"]

    dataset_name = test_info["dataset_name"]
    experiment_prefix = test_info["experiment_prefix"]
    num_repetitions = test_info.get("num_repetitions", 1)
    metadata_keys = test_info.get("metadata_keys", [])
    split_string = test_info.get("split", None)
    limit = test_info.get("limit", None)

    if split_string is None and limit is None and len(metadata_keys) == 0:
        return dataset_name, experiment_prefix, num_repetitions, metadata_keys
    limit = int(limit) if limit is not None else MAX_EXAMPLES_COUNT

    client = Client()
    examples = list(client.list_examples(dataset_name=dataset_name))

    if split_string is None and limit > 0:
        return examples[:limit], experiment_prefix, num_repetitions, metadata_keys

    splits = set(split_string.split(" "))

    test_examples = [
        example
        for example in examples
        if example.metadata is not None
        and example.metadata.get("dataset_split", None) is not None
        and bool(splits & set(example.metadata["dataset_split"]))
    ]

    test_examples = test_examples[:limit]

    if len(test_examples) > 0:
        return (test_examples, experiment_prefix, num_repetitions, metadata_keys)
    else:
        raise ValueError(f"No examples found for the dataset split: {split_string}")


def load_evaluators(config: dict[Any, Any]) -> tuple[Any, Any]:
    builtin_evaluators_config = config["tests"].get("assert", [])
    builtin_evaluators = generate_builtin_evaluator_functions(builtin_evaluators_config)

    evaluators_file_path = os.path.join(os.path.dirname(config_path), config["evaluators_file_path"])
    evaluators = load_function(evaluators_file_path, "evaluators") + builtin_evaluators
    summary_evaluators = load_function(evaluators_file_path, "summary_evaluators")

    return evaluators, summary_evaluators


async def run_evaluate(
    provider: dict[Any, Any],
    experiment_prefix: str,
    num_repetitions: int,
    metadatas: dict[str, Any] | None,
    **kwargs: dict[str, Any],
) -> tuple[Any, Any]:
    experiment_prefix_provider = experiment_prefix + provider["id"]
    prompt_func = load_run_function(config_path, config_file, provider)

    is_async = is_async_function(prompt_func)

    metadata = {"prompt_version": "1"}
    if metadatas:
        metadata.update(metadatas)

    common_args = {
        "experiment_prefix": experiment_prefix_provider,
        "num_repetitions": num_repetitions,
        "metadata": metadata,
        **kwargs,
    }

    if is_async:
        result = await aevaluate(prompt_func, **common_args)
        dataset_id = await result._manager.get_dataset_id()
    else:
        result = evaluate(prompt_func, **common_args)
        dataset_id = result._manager.dataset_id
    experiment_id = None
    if result._manager and result._manager._experiment and result._manager._experiment.id is not None:
        experiment_id = result._manager._experiment.id.__str__()

    return dataset_id, experiment_id


def extract_metadata(dataset_examples: str | list[Any], metadata_keys: list[str]) -> dict[str, Any] | None:
    if (
        len(metadata_keys) == 0
        or isinstance(dataset_examples, str)
        or not dataset_examples
        or dataset_examples[0].metadata is None
    ):
        return None

    metadatas = {}
    for key in metadata_keys:
        if key in dataset_examples[0].metadata:
            metadatas[key] = dataset_examples[0].metadata.get(key, None)

    return metadatas if len(metadatas) > 0 else None


async def main(config_file: dict[Any, Any]) -> None:
    dataset_examples, experiment_prefix, num_repetitions, metadata_keys = load_dataset(config_file)
    metadatas = extract_metadata(dataset_examples, metadata_keys)
    evaluators, summary_evaluators = load_evaluators(config_file)
    max_concurrency = config_file["tests"].get("max_concurrency", None)
    providers = config_file["providers"]

    dataset_id = None
    experiment_ids = []

    tasks = [
        run_evaluate(
            provider,
            experiment_prefix,
            data=dataset_examples,
            evaluators=evaluators,
            summary_evaluators=summary_evaluators,
            max_concurrency=max_concurrency,
            num_repetitions=num_repetitions,
            metadatas=metadatas,
        )
        for provider in providers
    ]

    # Run all tasks concurrently using asyncio.gather
    results = await asyncio.gather(*tasks)

    # Unpack results and collect dataset and experiment IDs
    for _dataset_id, experiment_id in results:
        dataset_id = _dataset_id
        experiment_ids.append(experiment_id)

    # Print the final comparison URL if there are multiple providers
    if len(providers) > 1:
        seed_url = "https://smith.langchain.com/o/"
        experiment_id_query_str = "%2C".join(experiment_ids)

        url = (
            f"{seed_url}{LANGCHAIN_TENANT_ID}/datasets/{dataset_id}/compare?selectedSessions={experiment_id_query_str}"
        )
        print(url)


if __name__ == "__main__":
    config_path = sys.argv[1] if len(sys.argv) > 1 else "config.yml"
    config_file = load_config(config_path)

    asyncio.run(main(config_file))



# Copyright 2024 Gaudiy Inc.
#
# SPDX-License-Identifier: Apache-2.0

import importlib.util
import inspect
from collections.abc import Callable
from typing import Any


def is_async_function(func: Callable[..., Any]) -> bool:
    return inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(func)


def load_function(module_path: str, function_name: str) -> Any:
    """
    Dynamically load a function from a given module.
    """
    if not isinstance(module_path, str) or not module_path:
        raise ValueError("Invalid or empty module path.")
    if not isinstance(function_name, str) or not function_name:
        raise ValueError("Invalid or empty function name.")

    try:
        spec = importlib.util.spec_from_file_location("module", module_path)
        if spec is None:
            raise ImportError(f"Cannot find module specification for the path: {module_path}")

        module = importlib.util.module_from_spec(spec)
        if spec.loader is None:
            raise ImportError("Module loader is not available.")
        spec.loader.exec_module(module)

        if not hasattr(module, function_name):
            raise AttributeError(f"Function '{function_name}' not found in module '{module_path}'")
    except (ImportError, AttributeError) as error:
        print(f"Error: {error}")
        raise
    return getattr(module, function_name)
